# Copyright (c) Meta Platforms, Inc. and affiliates.

# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
import subprocess
import pprint
import json

from .dependency_check import DChecker
from .spec import update_spec, generate_spec

CHECKOUT_METADATA_FILE = "rust-update-set.metadata"
DIVIDING_LINE = "="*150

class RustUpdate():
    def __init__(self, crates, checkout_metadata, patch=False) -> None:
        self.crates = crates
        self.checkout_metadata = checkout_metadata
        self.patch_warning = []
        self.pkgs_update = []
        self.pkgs_new = []
        self.dchecker = DChecker()
        self.pkgs_incompatible = []
        self.compat_pkgs_found = []
        self.patch=patch

    def clone(self, crate, dest=""):
        """ Clone repository using fedpkg, return False if it fails """
        output = subprocess.run(["fedpkg", "clone", f"rust-{crate}", dest], capture_output=True)
        if output.returncode != 0:
            if dest == "":
                print(f"Fail to execute 'fedpkg clone rust-{crate}'.")
            else:
                print(f"Fail to execute 'fedpkg clone rust-{crate} {dest}'.")
            self.print_cmd_output(output)
            return False
        print(f"Clone to rust-{crate} {dest}")
        return True

    def get_current_version(self, crate):
        """ Get the current version from existing spec file"""
        os.chdir(f"{os.getcwd()}/rust-{crate}")
        version = ""
        with open(f"rust-{crate}.spec") as f:
            for line in f.readlines():
                if line.startswith("Version:"):
                    version = line[len("Version:        "):].replace("\n","")
                    break
        os.chdir(f"{os.getcwd()}/..")
        return version

    def checkout(self, crate, version, type_="existing"):
        """ Checkout the packages need to updated or created"""
        if f"rust-{crate}" in self.checkout_metadata["checkout"]: return

        # clone repository for existing pkgs or create new folder for new pkgs
        if type_ == "existing":
            if version == "":
                version = self.dchecker.get_latest_crate_version(crate)
            print(DIVIDING_LINE)
            print(f"Update rust-{crate} to {version}")
            print(DIVIDING_LINE)
            clone_success = self.clone(crate)
            if not clone_success: return
            old_version = self.get_current_version(crate)
        else:
            print(DIVIDING_LINE)
            print(f"Create new package rust-{crate} {version}")
            print(DIVIDING_LINE)
            output = subprocess.run(["mkdir", f"rust-{crate}"], capture_output=True)
            old_version = -1
        if type_ == "new":
            self.checkout_metadata["checkout"][f"rust-{crate}"] = {
                "version": version, 
                "status": "clone", 
                "type": type_, 
                "old_version": old_version,
                "bugzilla": -1 # bugzilla id for Package Review
            }
        else:
            self.checkout_metadata["checkout"][f"rust-{crate}"] = {
                "version": version, 
                "status": "clone", 
                "type": type_, 
                "old_version": old_version,
                "push": "" # keep track of the remote pushed: origin or FAS name
            }
        self.dump_metadata()

        # update spec file for existing pkgs or create new spec file for new pkgs
        os.chdir(f"{os.getcwd()}/rust-{crate}")
        if type_ == "existing": 
            if update_spec(crate, version, self.patch):
                self.patch_warning.append(crate)
        else: generate_spec(crate, version, self.patch)

        # upload new sources
        if type_ == "existing":
            output = subprocess.run(["fedpkg", "new-sources", f"{crate}-{version}.crate"], capture_output=True)
            if output.returncode == 0:
                print(f"Upload new sources {crate}-{version}.crate and update 'source' file.")
            else:
                self.print_cmd_output(output)

        # do dependency check and recursively check out old dependencies and missing dependencies
        print("Dependency check...")
        dependencies = sorted(self.dchecker.get_missing_old_dependencies(crate, version), key=lambda x:(x[3], x[0]))
        os.chdir(f"{os.getcwd()}/..")
        if dependencies == []:
            print('\033[92m'+f"      No need to update old dependencies or package new dependencies"+'\033[0m')

        # recursively checkout old dependencies or missing dependencies
        self.checkout_metadata["dependencies"][crate] = []
        for dependency in dependencies:
            name, latest_version, version, d_type = dependency
            if d_type == "new":
                print('\033[93m'+f"      Dependency missing: need to package dependency {name} {version}"+'\033[0m')
            else:
                print('\033[93m'+f"      Dependency too old: need to upgrade package rust-{name} from {latest_version} to {version}"+'\033[0m')
            self.checkout_metadata["dependencies"][crate].append(name)
        self.dump_metadata()

        for dependency in dependencies:
            name, latest_version, version, d_type = dependency
            self.checkout(name, version, d_type)

    def update(self):
        """ Update all crates from cli arguments """
        for crate, version in self.crates:
            self.checkout(crate, version)
        self.update_compatibility_check()
        self.print_summary()
    
    def update_branch(self, branch):
        """ Update release branch for all packages using 'git rebase rawhide' """
        print(DIVIDING_LINE)
        print(f"Update Release Branch {branch}")
        print(DIVIDING_LINE)
        for pkg in self.checkout_metadata["checkout"]:
            pkg_info = self.checkout_metadata["checkout"][pkg]
            if pkg_info["type"] == "existing":
                # git switch branch, git rebase
                print(f"Update branch {branch} for {pkg}")
                os.chdir(f"{os.getcwd()}/{pkg}")
                output = subprocess.run(["git", "switch", branch], capture_output=True)
                if output.returncode == 0:
                    print(f"      Successfully switch branch to {branch}")
                else:
                    print(f"      Failed to switch branch to {branch}")
                    self.print_cmd_output(output, prefix=6*" ")
                    os.chdir(f"{os.getcwd()}/..")
                    continue
                output = subprocess.run(["git", "rebase", "rawhide"], capture_output=True)
                if output.returncode == 0:
                    print(f"      Successfully rebase branch {branch} onto rawhide")
                else:
                    print(f"      Failed to rebase branch {branch} onto rawhide")
                    self.print_cmd_output(output, prefix=6*" ")
                subprocess.run(["git", "switch", "rawhide"], capture_output=True)
                os.chdir(f"{os.getcwd()}/..")

    def check(self, pkgs):
        """ Execute compatibility check """
        print(DIVIDING_LINE)
        print("Update Compatibility Check")
        print(DIVIDING_LINE)
        return self.dchecker.check_compatibility(pkgs)

    def check_cmd(self, crates):
        """ Execute command 'check' without checking out packages need to be updated """
        pkgs = {}
        for crate, version in crates:
            exist, latest_version = self.dchecker.check_exist(crate)
            if exist:
                if version == "":
                    version = self.dchecker.get_latest_crate_version(crate)
                pkgs[f"rust-{crate}"] = {
                    "type": "existing",
                    "version": version,
                    "old_version": latest_version,
                }
        self.check(pkgs)

    def update_compatibility_check(self):
        """ Compatibility check for all packages need to be updated """
        self.compat_pkgs_found, self.pkgs_incompatible = self.check(self.checkout_metadata["checkout"])
        if len(self.pkgs_incompatible) == 0:
            print("Safe to update the whole set")
            return
        if len(self.pkgs_incompatible) != 0:
            print(DIVIDING_LINE)
            print("Create compat crates for incompatible packages")
            print(DIVIDING_LINE)
        for pkg in self.pkgs_incompatible:
            self.create_compat_crate(pkg[5:], self.checkout_metadata["checkout"][pkg]["old_version"])

    def create_compat_crate(self, crate, version):
        """ Create compat crates with old versions """
        # clone to rust-{crate}-{old_version}
        if not self.clone(crate, dest=f"rust-{crate}{version}"):
            return
        # modify rust-{crate}.spec with suffix
        os.chdir(f"{os.getcwd()}/rust-{crate}{version}")
        name_line = f"Name:           rust-%{{crate}}{version}\n"
        with open(f"rust-{crate}.spec", "r+") as f:
            cont = f.readlines()
            for i, line in enumerate(cont):
                if line.startswith(f"Name:"):
                    cont[i] = name_line
                    break
            f.seek(0)
            f.writelines(cont)
        output = subprocess.run(["mv", f"rust-{crate}.spec", f"rust-{crate}{version}.spec"])
        if f"rust-{crate}{version}" not in self.checkout_metadata["compats"]:
            self.checkout_metadata["compats"].append(f"rust-{crate}{version}")
            self.dump_metadata()
        os.chdir(f"{os.getcwd()}/..")

    def print_cmd_output(self, output, prefix=""):
        """ Print cmd output with warning color"""
        out = output.stdout.decode().split("\n")[:-1]
        err = output.stderr.decode().split("\n")[:-1]
        for line in out:
            print('\033[93m'+prefix+line+'\033[0m')
        for line in err:
            print('\033[93m'+prefix+line+'\033[0m')
        return len(out)+len(err)

    def print_summary(self):
        """ Print checkout results including pakcages to be updated, new packages and packages with patch warnings """
        print(DIVIDING_LINE)
        print("Summary")
        print(DIVIDING_LINE)
        pkgs = self.checkout_metadata["checkout"]
        pkgs_updates = []
        pkgs_new = []
        for crate in pkgs:
            pkg_data = pkgs[crate]
            if pkg_data["type"] == "existing":
                pkgs_updates.append((crate, pkg_data["old_version"], pkg_data["version"]))
            else:
                pkgs_new.append((crate, pkg_data["version"]))
        if len(pkgs_updates) != 0:
            print("Packages to be updated:")
            for pkg, version, new_version in pkgs_updates:
                print(f"      {pkg} from {version} to {new_version}")
        if len(pkgs_new) != 0:
            print("Missing dependencies to be packaged:")
            for pkg, version in pkgs_new:
                print(f"      {pkg} {version}")
        if len(self.patch_warning) != 0:
            print("Patch Warnings: the following pkgs fails to patch old changes to new spec or succeed with warning, please check.")
            for pkg in self.patch_warning:
                print('\033[93m'+f"      {pkg}"+'\033[0m')
        if len(self.pkgs_incompatible) != 0:
            print("Compat crates created:")
            for pkg in self.pkgs_incompatible:
                print(f"      {pkg} {self.checkout_metadata['checkout'][pkg]['old_version']}")
        if len(self.compat_pkgs_found) != 0:
            print("Compat crates found:")
            for pkg, version in self.compat_pkgs_found:
                print(f"      {pkg} {version}")

    def dump_metadata(self):
        """ Dump all checkout related metadata into json file"""
        with open(CHECKOUT_METADATA_FILE, "w") as f:
            json.dump(self.checkout_metadata, f, indent=4)

if __name__ == "__main__":
    pass
